# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Self, cast

from keystoneauth1 import adapter

from openstack import _log
from openstack.baremetal.v1 import _common
from openstack import exceptions
from openstack import resource
from openstack import utils


_logger = _log.setup_logging('openstack')


class Introspection(resource.Resource):
    resources_key = 'introspection'
    base_path = '/introspection'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = False
    allow_delete = True
    allow_list = True

    create_opts = resource.CreateOpts(requires_id=True)
    create_returns_body = False

    #: Timestamp at which the introspection was finished.
    finished_at = resource.Body('finished_at')
    #: The last error message (if any).
    error = resource.Body('error')
    #: The UUID of the introspection (matches the node UUID).
    id = resource.Body('uuid', alternate_id=True)
    #: Whether introspection is finished.
    is_finished = resource.Body('finished', type=bool)
    #: A list of relative links, including the self and bookmark links.
    links = resource.Body('links', type=list)
    #: Timestamp at which the introspection was started.
    started_at = resource.Body('started_at')
    #: The current introspection state.
    state = resource.Body('state')

    def abort(self, session: adapter.Adapter) -> None:
        """Abort introspection.

        :param session: The session to use for making this request.
        """
        if self.is_finished:
            return

        session = self._get_session(session)

        version = self._get_microversion(session)
        request = self._prepare_request(requires_id=True)
        request.url = utils.urljoin(request.url, 'abort')
        response = session.post(
            request.url,
            headers=request.headers,
            microversion=version,
            retriable_status_codes=_common.RETRIABLE_STATUS_CODES,
        )
        msg = f"Failed to abort introspection for node {self.id}"
        exceptions.raise_from_response(response, error_message=msg)

    def get_data(
        self,
        session: adapter.Adapter,
        processed: bool = True,
    ) -> dict[str, object]:
        """Get introspection data.

        Note that the introspection data format is not stable and can vary
        from environment to environment.

        :param session: The session to use for making this request.
        :param processed: Whether to fetch the final processed data (the
            default) or the raw unprocessed data as received from the ramdisk.
        :returns: introspection data from the most recent successful run.
        """
        session = self._get_session(session)

        version = self._get_microversion(session) if processed else '1.17'
        request = self._prepare_request(requires_id=True)
        request.url = utils.urljoin(request.url, 'data')
        if not processed:
            request.url = utils.urljoin(request.url, 'unprocessed')
        response = session.get(
            request.url, headers=request.headers, microversion=version
        )
        msg = f"Failed to fetch introspection data for node {self.id}"
        exceptions.raise_from_response(response, error_message=msg)
        return cast(dict[str, object], response.json())

    def wait(
        self,
        session: adapter.Adapter,
        timeout: int | float | None = None,
        ignore_error: bool = False,
    ) -> Self:
        """Wait for the node to reach the expected state.

        :param session: The session to use for making this request.
        :param timeout: How much (in seconds) to wait for the introspection.
            The value of ``None`` (the default) means no client-side timeout.
        :param ignore_error: If ``True``, this call will raise an exception
            if the introspection reaches the ``error`` state. Otherwise the
            error state is considered successful and the call returns.
        :return: This :class:`Introspection` instance.
        :raises: :class:`~openstack.exceptions.ResourceFailure` if
            introspection fails and ``ignore_error`` is ``False``.
        :raises: :class:`~openstack.exceptions.ResourceTimeout` on timeout.
        """
        if self._check_state(ignore_error):
            return self

        for count in utils.iterate_timeout(
            timeout, f"Timeout waiting for introspection on node {self.id}"
        ):
            self.fetch(session)
            if self._check_state(ignore_error):
                return self

            _logger.debug(
                'Still waiting for introspection of node %(node)s, '
                'the current state is "%(state)s"',
                {'node': self.id, 'state': self.state},
            )

        assert False, 'unreachable'

    def _check_state(self, ignore_error: bool) -> bool:
        if self.state == 'error' and not ignore_error:
            raise exceptions.ResourceFailure(
                f"Introspection of node {self.id} failed: {self.error}"
            )
        return cast(bool, self.is_finished)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections.abc import MutableMapping
from typing import Any, Literal, Self, overload

import requests

from keystoneauth1 import adapter

from openstack.common import tag
from openstack import exceptions
from openstack.image import _download
from openstack import resource
from openstack import utils


class Image(resource.Resource, tag.TagMixin, _download.DownloadMixin):
    resources_key = 'images'
    base_path = '/images'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_patch = True
    commit_method = 'PATCH'

    # Store all unknown attributes under 'properties' in the object.
    # Remotely they would be still in the resource root
    _store_unknown_attrs_as_properties = True

    _query_mapping = resource.QueryParameters(
        "id",
        "name",
        "visibility",
        "member_status",
        "owner",
        "status",
        "size_min",
        "size_max",
        "protected",
        "is_hidden",
        "sort_key",
        "sort_dir",
        "sort",
        "tag",
        "created_at",
        "updated_at",
        is_hidden="os_hidden",
    )

    # NOTE: Do not add "self" support here. If you've used Python before,
    # you know that self, while not being a reserved word, has special
    # meaning. You can't call a class initializer with the self name
    # as the first argument and then additionally in kwargs, as we
    # do when we're constructing instances from the JSON body.
    # Resource.list explicitly pops off any "self" keys from bodies so
    # that we don't end up getting the following:
    # TypeError: __init__() got multiple values for argument 'self'

    # The image data (bytes or a file-like object)
    data = None
    # Properties
    #: Hash of the image data used. The Image service uses this value
    #: for verification.
    checksum = resource.Body('checksum')
    #: The container format refers to whether the VM image is in a file
    #: format that also contains metadata about the actual VM.
    #: Container formats include OVF and Amazon AMI. In addition,
    #: a VM image might not have a container format - instead,
    #: the image is just a blob of unstructured data.
    container_format = resource.Body('container_format')
    #: The date and time when the image was created.
    created_at = resource.Body('created_at')
    #: Valid values are: aki, ari, ami, raw, iso, vhd, vdi, qcow2, or vmdk.
    #: The disk format of a VM image is the format of the underlying
    #: disk image. Virtual appliance vendors have different formats
    #: for laying out the information contained in a VM disk image.
    disk_format = resource.Body('disk_format')
    #: This field controls whether an image is displayed in the default
    #: image-list response
    is_hidden = resource.Body('os_hidden', type=bool)
    #: Defines whether the image can be deleted.
    #: *Type: bool*
    is_protected = resource.Body('protected', type=bool)
    #: The algorithm used to compute a secure hash of the image data
    #: for this image
    hash_algo = resource.Body('os_hash_algo')
    #: The hexdigest of the secure hash of the image data computed using
    #: the algorithm whose name is the value of the os_hash_algo property.
    hash_value = resource.Body('os_hash_value')
    #: The minimum disk size in GB that is required to boot the image.
    min_disk = resource.Body('min_disk')
    #: The minimum amount of RAM in MB that is required to boot the image.
    min_ram = resource.Body('min_ram')
    #: The name of the image.
    name = resource.Body('name')
    #: The ID of the owner, or project, of the image.
    owner = resource.Body('owner', alias='owner_id')
    #: The ID of the owner, or project, of the image. (backwards compat)
    owner_id = resource.Body('owner', alias='owner')
    # TODO(mordred) This is not how this works in v2. I mean, it's how it
    # should work, but it's not. We need to fix properties. They work right
    # in shade, so we can draw some logic from there.
    #: Properties, if any, that are associated with the image.
    properties = resource.Body('properties')
    #: The size of the image data, in bytes.
    size = resource.Body('size', type=int)
    #: When present, Glance will attempt to store the disk image data in the
    #: backing store indicated by the value of the header. When not present,
    #: Glance will store the disk image data in the backing store that is
    #: marked default. Valid values are: file, s3, rbd, swift, cinder,
    #: gridfs, sheepdog, or vsphere.
    store = resource.Body('store')
    #: The image status.
    status = resource.Body('status')
    #: The date and time when the image was updated.
    updated_at = resource.Body('updated_at')
    #: The virtual size of the image.
    virtual_size = resource.Body('virtual_size')
    #: The image visibility.
    visibility = resource.Body('visibility')
    #: The URL for the virtual machine image file.
    file = resource.Body('file')
    #: A list of URLs to access the image file in external store.
    #: This list appears if the show_multiple_locations option is set
    #: to true in the Image service's configuration file.
    locations = resource.Body('locations')
    #: The URL to access the image file kept in external store. It appears
    #: when you set the show_image_direct_url option to true in the
    #: Image service's configuration file.
    direct_url = resource.Body('direct_url')
    #: The URL to access the image file kept in external store.
    url = resource.Body('url')
    #: The location metadata.
    metadata = resource.Body('metadata', type=dict)

    # Additional Image Properties
    # https://docs.openstack.org/glance/latest/user/common-image-properties.html
    # http://docs.openstack.org/cli-reference/glance-property-keys.html
    #: The CPU architecture that must be supported by the hypervisor.
    architecture = resource.Body("architecture")
    #: The hypervisor type. Note that qemu is used for both QEMU and
    #: KVM hypervisor types.
    hypervisor_type = resource.Body("hypervisor_type")
    #: Optional property allows created servers to have a different bandwidth
    #: cap than that defined in the network they are attached to.
    instance_type_rxtx_factor = resource.Body(
        "instance_type_rxtx_factor",
        type=float,
    )
    # For snapshot images, this is the UUID of the server used to
    #: create this image.
    instance_uuid = resource.Body('instance_uuid')
    #: Specifies whether the image needs a config drive.
    #: `mandatory` or `optional` (default if property is not used).
    needs_config_drive = resource.Body('img_config_drive')
    #: The ID of an image stored in the Image service that should be used
    #: as the kernel when booting an AMI-style image.
    kernel_id = resource.Body('kernel_id')
    #: The common name of the operating system distribution in lowercase
    os_distro = resource.Body('os_distro')
    #: The operating system version as specified by the distributor.
    os_version = resource.Body('os_version')
    #: Secure Boot is a security standard. When the instance starts,
    #: Secure Boot first examines software such as firmware and OS by
    #: their signature and only allows them to run if the signatures are valid.
    needs_secure_boot = resource.Body('os_secure_boot')
    #: Time for graceful shutdown
    os_shutdown_timeout = resource.Body('os_shutdown_timeout', type=int)
    #: The ID of image stored in the Image service that should be used as
    #: the ramdisk when booting an AMI-style image.
    ramdisk_id = resource.Body('ramdisk_id')
    #: The virtual machine mode. This represents the host/guest ABI
    #: (application binary interface) used for the virtual machine.
    vm_mode = resource.Body('vm_mode')
    #: The preferred number of sockets to expose to the guest.
    hw_cpu_sockets = resource.Body('hw_cpu_sockets', type=int)
    #: The preferred number of cores to expose to the guest.
    hw_cpu_cores = resource.Body('hw_cpu_cores', type=int)
    #: The preferred number of threads to expose to the guest.
    hw_cpu_threads = resource.Body('hw_cpu_threads', type=int)
    #: Specifies the type of disk controller to attach disk devices to.
    #: One of scsi, virtio, uml, xen, ide, or usb.
    hw_disk_bus = resource.Body('hw_disk_bus')
    #: Used to pin the virtual CPUs (vCPUs) of instances to the
    #: host's physical CPU cores (pCPUs).
    hw_cpu_policy = resource.Body('hw_cpu_policy')
    #: Defines how hardware CPU threads in a simultaneous
    #: multithreading-based (SMT) architecture be used.
    hw_cpu_thread_policy = resource.Body('hw_cpu_thread_policy')
    #: Adds a random-number generator device to the image's instances.
    hw_rng_model = resource.Body('hw_rng_model')
    #: For libvirt: Enables booting an ARM system using the specified
    #: machine type.
    #: For Hyper-V: Specifies whether the Hyper-V instance will be a
    #: generation 1 or generation 2 VM.
    hw_machine_type = resource.Body('hw_machine_type')
    #: Enables the use of VirtIO SCSI (virtio-scsi) to provide block device
    #: access for compute instances; by default, instances use VirtIO Block
    #: (virtio-blk).
    hw_scsi_model = resource.Body('hw_scsi_model')
    #: Specifies the count of serial ports that should be provided.
    hw_serial_port_count = resource.Body('hw_serial_port_count', type=int)
    #: The video image driver used.
    hw_video_model = resource.Body('hw_video_model')
    #: Maximum RAM for the video image.
    hw_video_ram = resource.Body('hw_video_ram', type=int)
    #: Enables a virtual hardware watchdog device that carries out the
    #: specified action if the server hangs.
    hw_watchdog_action = resource.Body('hw_watchdog_action')
    #: The kernel command line to be used by the libvirt driver, instead
    #: of the default.
    os_command_line = resource.Body('os_command_line')
    #: Specifies the model of virtual network interface device to use.
    hw_vif_model = resource.Body('hw_vif_model')
    #: If true, this enables the virtio-net multiqueue feature.
    #: In this case, the driver sets the number of queues equal to the
    #: number of guest vCPUs. This makes the network performance scale
    #: across a number of vCPUs.
    is_hw_vif_multiqueue_enabled = resource.Body(
        'hw_vif_multiqueue_enabled',
        type=bool,
    )
    #: If true, enables the BIOS bootmenu.
    is_hw_boot_menu_enabled = resource.Body('hw_boot_menu', type=bool)
    #: The virtual SCSI or IDE controller used by the hypervisor.
    vmware_adaptertype = resource.Body('vmware_adaptertype')
    #: A VMware GuestID which describes the operating system installed
    #: in the image.
    vmware_ostype = resource.Body('vmware_ostype')
    #: If true, the root partition on the disk is automatically resized
    #: before the instance boots.
    has_auto_disk_config = resource.Body('auto_disk_config')
    #: The operating system installed on the image.
    os_type = resource.Body('os_type')
    #: The operating system admin username.
    os_admin_user = resource.Body('os_admin_user')
    #: A string boolean, which if "true", QEMU guest agent will be exposed
    #: to the instance.
    hw_qemu_guest_agent = resource.Body('hw_qemu_guest_agent', type=str)
    #: If true, require quiesce on snapshot via QEMU guest agent.
    os_require_quiesce = resource.Body('os_require_quiesce', type=bool)
    #: The URL for the schema describing a virtual machine image.
    schema = resource.Body('schema')

    def _action(
        self, session: adapter.Adapter, action: str
    ) -> requests.Response:
        """Call an action on an image ID."""
        url = utils.urljoin(self.base_path, self.id, 'actions', action)
        return session.post(url)

    def deactivate(self, session: adapter.Adapter) -> None:
        """Deactivate an image

        Note: Only administrative users can view image locations
        for deactivated images.
        """
        self._action(session, "deactivate")

    def reactivate(self, session: adapter.Adapter) -> None:
        """Reactivate an image

        Note: The image must exist in order to be reactivated.
        """
        self._action(session, "reactivate")

    def upload(
        self, session: adapter.Adapter, *, data: Any = None
    ) -> requests.Response:
        """Upload data into an existing image

        :param session: The session to use for making this request
        :param data: Optional data to be uploaded. If not provided, the
            `~Image.data` attribute will be used
        :returns: The server response
        """
        if data:
            self.data = data
        url = utils.urljoin(self.base_path, self.id, 'file')
        return session.put(
            url,
            data=self.data,
            headers={"Content-Type": "application/octet-stream", "Accept": ""},
        )

    def stage(self, session: adapter.Adapter, *, data: Any = None) -> Self:
        """Stage binary image data into an existing image

        :param session: The session to use for making this request
        :param data: Optional data to be uploaded. If not provided, the
            `~Image.data` attribute will be used
        :returns: The server response
        """
        if data:
            self.data = data

        url = utils.urljoin(self.base_path, self.id, 'stage')
        response = session.put(
            url,
            data=self.data,
            headers={"Content-Type": "application/octet-stream", "Accept": ""},
        )
        self._translate_response(response, has_body=False)
        return self

    def import_image(
        self,
        session: adapter.Adapter,
        method: str = 'glance-direct',
        *,
        uri: str | None = None,
        remote_region: str | None = None,
        remote_image_id: str | None = None,
        remote_service_interface: str | None = None,
        store: Any = None,
        stores: list[Any] | None = None,
        all_stores: bool | None = None,
        all_stores_must_succeed: bool | None = None,
    ) -> requests.Response:
        """Import Image via interoperable image import process"""
        if all_stores and (store or stores):
            raise exceptions.InvalidRequest(
                'all_stores is mutually exclusive with store and stores'
            )
        if store and stores:
            raise exceptions.InvalidRequest(
                'store and stores are mutually exclusive. stores should be '
                'preferred.'
            )
        if store:
            stores = [store]
        else:
            stores = stores or []

        url = utils.urljoin(self.base_path, self.id, 'import')
        data: dict[str, Any] = {'method': {'name': method}}

        if uri:
            if method != 'web-download':
                raise exceptions.InvalidRequest(
                    'URI is only supported with method: "web-download"'
                )
            data['method']['uri'] = uri

        if remote_region and remote_image_id:
            if remote_service_interface:
                data['method']['glance_service_interface'] = (
                    remote_service_interface
                )
                data['method']['glance_region'] = remote_region
                data['method']['glance_image_id'] = remote_image_id

        if all_stores is not None:
            data['all_stores'] = all_stores
        if all_stores_must_succeed is not None:
            data['all_stores_must_succeed'] = all_stores_must_succeed
        if stores:
            data['stores'] = [s.id for s in stores]

        headers = {}
        # Backward compat
        if store is not None:
            headers = {'X-Image-Meta-Store': store.id}

        return session.post(url, json=data, headers=headers)

    def _consume_header_attrs(
        self, attrs: MutableMapping[str, Any]
    ) -> dict[str, Any]:
        self.image_import_methods = []
        _image_import_methods = attrs.pop('OpenStack-image-import-methods', '')
        if _image_import_methods:
            self.image_import_methods = _image_import_methods.split(',')

        return super()._consume_header_attrs(attrs)

    def _prepare_request(
        self,
        requires_id: bool | None = None,
        prepend_key: bool = False,
        patch: bool = False,
        base_path: str | None = None,
        params: Any = None,
        **kwargs: Any,
    ) -> resource._Request:
        request = super()._prepare_request(
            requires_id=requires_id,
            prepend_key=prepend_key,
            patch=patch,
            base_path=base_path,
            params=params,
            **kwargs,
        )
        if patch:
            headers = {
                'Content-Type': 'application/openstack-images-v2.1-json-patch',
                'Accept': '',
            }
            request.headers.update(headers)

        return request

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: Literal[False],
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None: ...

    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None:
        # Do a regular search first (ignoring missing)
        result = super().find(
            session,
            name_or_id,
            ignore_missing=True,
            list_base_path=list_base_path,
            microversion=microversion,
            all_projects=all_projects,
            **params,
        )

        if result:
            return result
        else:
            # Search also in hidden images
            params['is_hidden'] = True
            data = cls.list(session, **params)

            result = cls._get_one_match(name_or_id, data)
            if result is not None:
                return result

        if ignore_missing:
            return None
        raise exceptions.NotFoundException(
            f"No {cls.__name__} found for {name_or_id}"
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import uuid

from openstack.message.v2 import message
from openstack.tests.unit import base

FAKE1 = {
    'age': 456,
    'body': {
        'current_bytes': '0',
        'event': 'BackupProgress',
        'total_bytes': '99614720',
    },
    'id': '578ee000508f153f256f717d',
    'href': '/v2/queues/queue1/messages/578ee000508f153f256f717d',
    'ttl': 3600,
    'queue_name': 'queue1',
}


FAKE2 = {
    'age': 456,
    'body': {
        'current_bytes': '0',
        'event': 'BackupProgress',
        'total_bytes': '99614720',
    },
    'id': '578ee000508f153f256f717d',
    'href': '/v2/queues/queue1/messages/578ee000508f153f256f717d',
    'ttl': 3600,
    'queue_name': 'queue1',
    'client_id': 'OLD_CLIENT_ID',
    'project_id': 'OLD_PROJECT_ID',
}


class TestMessage(base.TestCase):
    def test_basic(self):
        sot = message.Message()
        self.assertEqual('messages', sot.resources_key)
        self.assertEqual('/queues/%(queue_name)s/messages', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = message.Message.new(**FAKE2)
        self.assertEqual(FAKE2['age'], sot.age)
        self.assertEqual(FAKE2['body'], sot.body)
        self.assertEqual(FAKE2['id'], sot.id)
        self.assertEqual(FAKE2['href'], sot.href)
        self.assertEqual(FAKE2['ttl'], sot.ttl)
        self.assertEqual(FAKE2['queue_name'], sot.queue_name)
        self.assertEqual(FAKE2['client_id'], sot.client_id)
        self.assertEqual(FAKE2['project_id'], sot.project_id)

    @mock.patch.object(uuid, 'uuid4')
    def test_post(self, mock_uuid):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.post.return_value = resp
        resources = [
            '/v2/queues/queue1/messages/578ee000508f153f256f717d'
            '/v2/queues/queue1/messages/579edd6c368cb61de9a7e233'
        ]
        resp.json.return_value = {'resources': resources}
        sess.get_project_id.return_value = 'NEW_PROJECT_ID'
        mock_uuid.return_value = 'NEW_CLIENT_ID'
        messages = [
            {'body': {'key': 'value1'}, 'ttl': 3600},
            {'body': {'key': 'value2'}, 'ttl': 1800},
        ]

        sot = message.Message(**FAKE1)
        res = sot.post(sess, messages)

        url = '/queues/{queue}/messages'.format(queue=FAKE1['queue_name'])
        headers = {
            'Client-ID': 'NEW_CLIENT_ID',
            'X-PROJECT-ID': 'NEW_PROJECT_ID',
        }
        sess.post.assert_called_once_with(
            url, headers=headers, json={'messages': messages}
        )
        sess.get_project_id.assert_called_once_with()
        resp.json.assert_called_once_with()
        self.assertEqual(resources, res)

    def test_post_client_id_project_id_exist(self):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.post.return_value = resp
        resources = [
            '/v2/queues/queue1/messages/578ee000508f153f256f717d'
            '/v2/queues/queue1/messages/579edd6c368cb61de9a7e233'
        ]
        resp.json.return_value = {'resources': resources}
        messages = [
            {'body': {'key': 'value1'}, 'ttl': 3600},
            {'body': {'key': 'value2'}, 'ttl': 1800},
        ]

        sot = message.Message(**FAKE2)
        res = sot.post(sess, messages)

        url = '/queues/{queue}/messages'.format(queue=FAKE2['queue_name'])
        headers = {
            'Client-ID': 'OLD_CLIENT_ID',
            'X-PROJECT-ID': 'OLD_PROJECT_ID',
        }
        sess.post.assert_called_once_with(
            url, headers=headers, json={'messages': messages}
        )
        resp.json.assert_called_once_with()
        self.assertEqual(resources, res)

    @mock.patch.object(uuid, 'uuid4')
    def test_get(self, mock_uuid):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.get.return_value = resp
        sess.get_project_id.return_value = 'NEW_PROJECT_ID'
        mock_uuid.return_value = 'NEW_CLIENT_ID'

        sot = message.Message(**FAKE1)
        sot._translate_response = mock.Mock()
        res = sot.fetch(sess)

        url = 'queues/{queue}/messages/{message}'.format(
            queue=FAKE1['queue_name'],
            message=FAKE1['id'],
        )
        headers = {
            'Client-ID': 'NEW_CLIENT_ID',
            'X-PROJECT-ID': 'NEW_PROJECT_ID',
        }
        sess.get.assert_called_with(url, headers=headers, skip_cache=False)
        sess.get_project_id.assert_called_once_with()
        sot._translate_response.assert_called_once_with(resp)
        self.assertEqual(sot, res)

    def test_get_client_id_project_id_exist(self):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.get.return_value = resp

        sot = message.Message(**FAKE1)
        sot._translate_response = mock.Mock()
        res = sot.fetch(sess)

        url = 'queues/{queue}/messages/{message}'.format(
            queue=FAKE2['queue_name'],
            message=FAKE2['id'],
        )
        sot = message.Message(**FAKE2)
        sot._translate_response = mock.Mock()
        res = sot.fetch(sess)
        headers = {
            'Client-ID': 'OLD_CLIENT_ID',
            'X-PROJECT-ID': 'OLD_PROJECT_ID',
        }
        sess.get.assert_called_with(url, headers=headers, skip_cache=False)
        sot._translate_response.assert_called_once_with(resp)
        self.assertEqual(sot, res)

    @mock.patch.object(uuid, 'uuid4')
    def test_delete_unclaimed(self, mock_uuid):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.delete.return_value = resp
        sess.get_project_id.return_value = 'NEW_PROJECT_ID'
        mock_uuid.return_value = 'NEW_CLIENT_ID'

        sot = message.Message(**FAKE1)
        sot.claim_id = None
        sot._translate_response = mock.Mock()
        sot.delete(sess)

        url = 'queues/{queue}/messages/{message}'.format(
            queue=FAKE1['queue_name'],
            message=FAKE1['id'],
        )
        headers = {
            'Client-ID': 'NEW_CLIENT_ID',
            'X-PROJECT-ID': 'NEW_PROJECT_ID',
        }
        sess.delete.assert_called_with(url, headers=headers)
        sess.get_project_id.assert_called_once_with()
        sot._translate_response.assert_called_once_with(resp, has_body=False)

    @mock.patch.object(uuid, 'uuid4')
    def test_delete_claimed(self, mock_uuid):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.delete.return_value = resp
        sess.get_project_id.return_value = 'NEW_PROJECT_ID'
        mock_uuid.return_value = 'NEW_CLIENT_ID'

        sot = message.Message(**FAKE1)
        sot.claim_id = 'CLAIM_ID'
        sot._translate_response = mock.Mock()
        sot.delete(sess)

        url = 'queues/{queue}/messages/{message}?claim_id={cid}'.format(
            queue=FAKE1['queue_name'],
            message=FAKE1['id'],
            cid='CLAIM_ID',
        )
        headers = {
            'Client-ID': 'NEW_CLIENT_ID',
            'X-PROJECT-ID': 'NEW_PROJECT_ID',
        }
        sess.delete.assert_called_with(url, headers=headers)
        sess.get_project_id.assert_called_once_with()
        sot._translate_response.assert_called_once_with(resp, has_body=False)

    def test_delete_client_id_project_id_exist(self):
        sess = mock.Mock()
        resp = mock.Mock()
        sess.delete.return_value = resp

        sot = message.Message(**FAKE2)
        sot.claim_id = None
        sot._translate_response = mock.Mock()
        sot.delete(sess)

        url = 'queues/{queue}/messages/{message}'.format(
            queue=FAKE2['queue_name'],
            message=FAKE2['id'],
        )
        headers = {
            'Client-ID': 'OLD_CLIENT_ID',
            'X-PROJECT-ID': 'OLD_PROJECT_ID',
        }
        sess.delete.assert_called_with(url, headers=headers)
        sot._translate_response.assert_called_once_with(resp, has_body=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import security_groups_default_statefulness as sgds
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'id': IDENTIFIER,
    'project_id': 'fb0e16da8ce7435c8a44464773e21660',
    'stateful': False,
}


class TestSecurityGroupsDefaultStatefulness(base.TestCase):
    def test_basic(self):
        sot = sgds.SecurityGroupsDefaultStatefulness()
        self.assertEqual(
            'security_group_default_statefulness', sot.resource_key
        )
        self.assertEqual(
            'security_groups_default_statefulness', sot.resources_key
        )
        self.assertEqual(
            '/security-groups-default-statefulness', sot.base_path
        )
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                'id': 'id',
                'limit': 'limit',
                'marker': 'marker',
                'project_id': 'project_id',
                'sort_dir': 'sort_dir',
                'sort_key': 'sort_key',
                'stateful': 'stateful',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = sgds.SecurityGroupsDefaultStatefulness(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['stateful'], sot.stateful)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.orchestration.v1 import stack_event
from openstack.tests.unit import base


FAKE_ID = 'ce8ae86c-9810-4cb1-8888-7fb53bc523bf'
FAKE_NAME = 'test_stack'
FAKE = {
    'event_time': '2015-03-09T12:15:57.233772',
    'id': FAKE_ID,
    'links': [{'href': f'stacks/{FAKE_NAME}/{FAKE_ID}', 'rel': 'self'}],
    'logical_resource_id': 'my_test_group',
    'physical_resource_id': 'my_test_group',
    'resource_name': 'my_test_resource',
    'resource_status': 'CREATE_IN_PROGRESS',
    'resource_status_reason': 'state changed',
}


class TestStackEvent(base.TestCase):
    def test_basic(self):
        sot = stack_event.StackEvent()
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = stack_event.StackEvent(**FAKE)
        self.assertEqual(FAKE['event_time'], sot.event_time)
        self.assertEqual(FAKE['id'], sot.id)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['logical_resource_id'], sot.logical_resource_id)
        self.assertEqual(
            FAKE['physical_resource_id'], sot.physical_resource_id
        )
        self.assertEqual(FAKE['resource_name'], sot.resource_name)
        self.assertEqual(FAKE['resource_status'], sot.resource_status)
        self.assertEqual(
            FAKE['resource_status_reason'], sot.resource_status_reason
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings

import pbr.version

warnings.warn(
    f'The {__name__} module is deprecated. Prefer use of importlib.metadata '
    f'to inspect version information for packages.',
    DeprecationWarning,
)

__version__ = pbr.version.VersionInfo('openstacksdk').version_string()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# needed to prevent circular imports with openstack.network.v2.agent
from __future__ import annotations

from typing import Any

from keystoneauth1 import adapter

from openstack import exceptions
from openstack.network.v2 import agent as _agent
from openstack import resource
from openstack import utils


class BgpSpeaker(resource.Resource):
    resource_key = 'bgp_speaker'
    resources_key = 'bgp_speakers'
    base_path = '/bgp-speakers'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: The Id of the BGP Speaker
    id = resource.Body('id')
    #: The BGP speaker's name.
    name = resource.Body('name')
    #: The ID of the project that owns the BGP Speaker.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The IP version (4 or 6) of the BGP Speaker.
    ip_version = resource.Body('ip_version')
    #: Whether to enable or disable the advertisement of floating ip host
    #: routes by the BGP Speaker. True by default.
    advertise_floating_ip_host_routes = resource.Body(
        'advertise_floating_ip_host_routes'
    )
    #: Whether to enable or disable the advertisement of tenant network
    #: routes by the BGP Speaker. True by default.
    advertise_tenant_networks = resource.Body('advertise_tenant_networks')
    #: The local Autonomous System number of the BGP Speaker.
    local_as = resource.Body('local_as')
    #: A list of network IDs to which the BGP Speaker is associated.
    networks = resource.Body('networks', type=list)

    def _put(
        self, session: adapter.Adapter, url: str, body: dict[str, Any]
    ) -> Any:
        resp = session.put(url, json=body)
        exceptions.raise_from_response(resp)
        return resp

    def add_bgp_peer(
        self, session: adapter.Adapter, peer_id: str
    ) -> dict[str, Any]:
        """Add BGP Peer to a BGP Speaker

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param peer_id: id of the peer to associate with the speaker.

        :returns: A dictionary as the API Reference describes it.

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'add_bgp_peer')
        body = {'bgp_peer_id': peer_id}
        resp = self._put(session, url, body)
        return resp.json()  # type: ignore[no-any-return]

    def remove_bgp_peer(self, session: adapter.Adapter, peer_id: str) -> None:
        """Remove BGP Peer from a BGP Speaker

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param peer_id: The ID of the peer to disassociate from the speaker.

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'remove_bgp_peer')
        body = {'bgp_peer_id': peer_id}
        self._put(session, url, body)

    def add_gateway_network(
        self, session: adapter.Adapter, network_id: str
    ) -> dict[str, Any]:
        """Add Network to a BGP Speaker

        :param: session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param network_id: The ID of the network to associate with the speaker

        :returns: A dictionary as the API Reference describes it.
        """
        body = {'network_id': network_id}
        url = utils.urljoin(self.base_path, self.id, 'add_gateway_network')
        resp = session.put(url, json=body)
        return resp.json()  # type: ignore[no-any-return]

    def remove_gateway_network(
        self, session: adapter.Adapter, network_id: str
    ) -> None:
        """Delete Network from a BGP Speaker

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param network_id: The ID of the network to disassociate
               from the speaker
        """
        body = {'network_id': network_id}
        url = utils.urljoin(self.base_path, self.id, 'remove_gateway_network')
        session.put(url, json=body)

    def get_advertised_routes(
        self, session: adapter.Adapter
    ) -> dict[str, Any]:
        """List routes advertised by a BGP Speaker

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :returns: The response as a list of routes (cidr/nexthop pair
                  advertised by the BGP Speaker.

        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'get_advertised_routes')
        resp = session.get(url)
        exceptions.raise_from_response(resp)
        self._body.attributes.update(resp.json())
        return resp.json()  # type: ignore[no-any-return]

    def get_bgp_dragents(self, session: adapter.Adapter) -> list[_agent.Agent]:
        """List Dynamic Routing Agents hosting a specific BGP Speaker

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :returns: The response as a list of dragents hosting a specific
                  BGP Speaker.
        :rtype: :class:`~openstack.network.v2.agent.Agent`
        :raises: :class:`~openstack.exceptions.SDKException` on error.
        """
        url = utils.urljoin(self.base_path, self.id, 'bgp-dragents')
        resp = session.get(url)
        exceptions.raise_from_response(resp)
        self._body.attributes.update(resp.json())
        agent_ids = [ag['id'] for ag in resp.json()['agents']]
        agents = _agent.Agent.list(session=session)
        return [ag for ag in agents if ag.id in agent_ids]

    def add_bgp_speaker_to_dragent(
        self, session: adapter.Adapter, bgp_agent_id: str
    ) -> None:
        """Add BGP Speaker to a Dynamic Routing Agent

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param bgp_agent_id: The id of the dynamic routing agent to which
                             add the speaker.
        """
        body = {'bgp_speaker_id': self.id}
        url = utils.urljoin('agents', bgp_agent_id, 'bgp-drinstances')
        session.post(url, json=body)

    def remove_bgp_speaker_from_dragent(
        self, session: adapter.Adapter, bgp_agent_id: str
    ) -> None:
        """Delete BGP Speaker from a Dynamic Routing Agent

        :param session: The session to communicate through.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param bgp_agent_id: The id of the dynamic routing agent from which
                             remove the speaker.
        """
        url = utils.urljoin('agents', bgp_agent_id, 'bgp-drinstances', self.id)
        session.delete(url)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystoneauth1 import exceptions as ks_exc
import requests.exceptions

from openstack.config import cloud_region
from openstack import connection
from openstack import exceptions
from openstack.tests.unit import base


class TestFromConf(base.TestCase):
    def _get_conn(self, **from_conf_kwargs):
        oslocfg = self._load_ks_cfg_opts()
        # Throw name in here to prove **kwargs is working
        config = cloud_region.from_conf(
            oslocfg,
            session=self.cloud.session,
            name='from_conf.example.com',
            **from_conf_kwargs,
        )
        self.assertEqual('from_conf.example.com', config.name)

        return connection.Connection(config=config, strict_proxies=True)

    def test_adapter_opts_set(self):
        """Adapter opts specified in the conf."""
        conn = self._get_conn()

        discovery = {
            "versions": {
                "values": [
                    {
                        "status": "stable",
                        "updated": "2019-06-01T00:00:00Z",
                        "media-types": [
                            {
                                "base": "application/json",
                                "type": "application/vnd.openstack.heat-v2+json",  # noqa: E501
                            }
                        ],
                        "id": "v2.0",
                        "links": [
                            {
                                "href": "https://example.org:8888/heat/v2",
                                "rel": "self",
                            }
                        ],
                    }
                ]
            }
        }
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://example.org:8888/heat/v2',
                    json=discovery,
                ),
                dict(
                    method='GET',
                    uri='https://example.org:8888/heat/v2/foo',
                    json={'foo': {}},
                ),
            ]
        )

        adap = conn.orchestration
        self.assertEqual('SpecialRegion', adap.region_name)
        self.assertEqual('orchestration', adap.service_type)
        self.assertEqual('internal', adap.interface)
        self.assertEqual(
            'https://example.org:8888/heat/v2', adap.endpoint_override
        )

        adap.get('/foo')
        self.assert_calls()

    def test_default_adapter_opts(self):
        """Adapter opts are registered, but all defaulting in conf."""
        conn = self._get_conn()

        server_id = str(uuid.uuid4())
        server_name = self.getUniqueString('name')
        fake_server = base.make_fake_server(server_id, server_name)

        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', 'detail']
                    ),
                    json={'servers': [fake_server]},
                ),
            ]
        )

        # Nova has empty adapter config, so these default
        adap = conn.compute
        self.assertIsNone(adap.region_name)
        self.assertEqual('compute', adap.service_type)
        self.assertEqual('public', adap.interface)
        self.assertIsNone(adap.endpoint_override)

        s = next(adap.servers())
        self.assertEqual(s.id, server_id)
        self.assertEqual(s.name, server_name)
        self.assert_calls()

    def test_service_not_ready_catalog(self):
        """Adapter opts are registered, but all defaulting in conf."""
        conn = self._get_conn()

        server_id = str(uuid.uuid4())
        server_name = self.getUniqueString('name')
        fake_server = base.make_fake_server(server_id, server_name)

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://compute.example.com/v2.1/',
                    exc=requests.exceptions.ConnectionError,
                ),
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['servers', 'detail']
                    ),
                    json={'servers': [fake_server]},
                ),
            ]
        )

        self.assertRaises(
            exceptions.ServiceDiscoveryException, getattr, conn, 'compute'
        )

        # Nova has empty adapter config, so these default
        adap = conn.compute
        self.assertIsNone(adap.region_name)
        self.assertEqual('compute', adap.service_type)
        self.assertEqual('public', adap.interface)
        self.assertIsNone(adap.endpoint_override)

        s = next(adap.servers())
        self.assertEqual(s.id, server_id)
        self.assertEqual(s.name, server_name)
        self.assert_calls()

    def test_name_with_dashes(self):
        conn = self._get_conn()

        discovery = {
            "versions": {
                "values": [
                    {
                        "status": "stable",
                        "id": "v1",
                        "links": [
                            {
                                "href": "https://example.org:5050/v1",
                                "rel": "self",
                            }
                        ],
                    }
                ]
            }
        }
        status = {'finished': True, 'error': None}
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://example.org:5050',
                    json=discovery,
                ),
                # strict-proxies means we're going to fetch the discovery
                # doc from the versioned endpoint to verify it works.
                dict(
                    method='GET',
                    uri='https://example.org:5050/v1',
                    json=discovery,
                ),
                dict(
                    method='GET',
                    uri='https://example.org:5050/v1/introspection/abcd',
                    json=status,
                ),
            ]
        )

        adap = conn.baremetal_introspection
        self.assertEqual('baremetal-introspection', adap.service_type)
        self.assertEqual('public', adap.interface)
        self.assertEqual('https://example.org:5050/v1', adap.endpoint_override)

        self.assertTrue(adap.get_introspection('abcd').is_finished)

    def test_service_not_ready_endpoint_override(self):
        conn = self._get_conn()

        discovery = {
            "versions": {
                "values": [
                    {
                        "status": "stable",
                        "id": "v1",
                        "links": [
                            {
                                "href": "https://example.org:5050/v1",
                                "rel": "self",
                            }
                        ],
                    }
                ]
            }
        }
        status = {'finished': True, 'error': None}
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://example.org:5050',
                    exc=requests.exceptions.ConnectTimeout,
                ),
                dict(
                    method='GET',
                    uri='https://example.org:5050',
                    json=discovery,
                ),
                # strict-proxies means we're going to fetch the discovery
                # doc from the versioned endpoint to verify it works.
                dict(
                    method='GET',
                    uri='https://example.org:5050/v1',
                    json=discovery,
                ),
                dict(
                    method='GET',
                    uri='https://example.org:5050/v1/introspection/abcd',
                    json=status,
                ),
            ]
        )

        self.assertRaises(
            exceptions.ServiceDiscoveryException,
            getattr,
            conn,
            'baremetal_introspection',
        )

        adap = conn.baremetal_introspection
        self.assertEqual('baremetal-introspection', adap.service_type)
        self.assertEqual('public', adap.interface)
        self.assertEqual('https://example.org:5050/v1', adap.endpoint_override)

        self.assertTrue(adap.get_introspection('abcd').is_finished)

    def assert_service_disabled(
        self, service_type, expected_reason, **from_conf_kwargs
    ):
        conn = self._get_conn(**from_conf_kwargs)
        # The _ServiceDisabledProxyShim loads up okay...
        adap = getattr(conn, service_type)
        # ...but freaks out if you try to use it.
        ex = self.assertRaises(
            exceptions.ServiceDisabledException, getattr, adap, 'get'
        )
        self.assertIn(
            f"Service '{service_type}' is disabled because its configuration "
            "could not be loaded.",
            ex.message,
        )
        self.assertIn(expected_reason, ex.message)

    def test_no_such_conf_section(self):
        """No conf section (therefore no adapter opts) for service type."""
        del self.oslo_config_dict['heat']
        self.assert_service_disabled(
            'orchestration',
            "No section for project 'heat' (service type 'orchestration') was "
            "present in the config.",
        )

    def test_no_such_conf_section_ignore_service_type(self):
        """Ignore absent conf section if service type not requested."""
        del self.oslo_config_dict['heat']
        self.assert_service_disabled(
            'orchestration',
            "Not in the list of requested service_types.",
            # 'orchestration' absent from this list
            service_types=['compute'],
        )

    def test_no_adapter_opts(self):
        """Conf section present, but opts for service type not registered."""
        self.oslo_config_dict['heat'] = None
        self.assert_service_disabled(
            'orchestration',
            "Encountered an exception attempting to process config for "
            "project 'heat' (service type 'orchestration'): no such option",
        )

    def test_no_adapter_opts_ignore_service_type(self):
        """Ignore unregistered conf section if service type not requested."""
        self.oslo_config_dict['heat'] = None
        self.assert_service_disabled(
            'orchestration',
            "Not in the list of requested service_types.",
            # 'orchestration' absent from this list
            service_types=['compute'],
        )

    def test_invalid_adapter_opts(self):
        """Adapter opts are bogus, in exception-raising ways."""
        self.oslo_config_dict['heat'] = {
            'interface': 'public',
            'valid_interfaces': 'private',
        }
        self.assert_service_disabled(
            'orchestration',
            "Encountered an exception attempting to process config for "
            "project 'heat' (service type 'orchestration'): interface and "
            "valid_interfaces are mutually exclusive.",
        )

    def test_no_session(self):
        # TODO(efried): Currently calling without a Session is not implemented.
        self.assertRaises(
            exceptions.ConfigException,
            cloud_region.from_conf,
            self._load_ks_cfg_opts(),
        )

    def test_no_endpoint(self):
        """Conf contains adapter opts, but service type not in catalog."""
        self.os_fixture.v3_token.remove_service('monitoring')
        conn = self._get_conn()
        # Monasca is not in the service catalog
        self.assertRaises(
            ks_exc.catalog.EndpointNotFound, getattr, conn, 'monitoring'
        )

    def test_no_endpoint_ignore_service_type(self):
        """Bogus service type disabled if not in requested service_types."""
        self.assert_service_disabled(
            'monitoring',
            "Not in the list of requested service_types.",
            # 'monitoring' absent from this list
            service_types={'compute', 'orchestration', 'bogus'},
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime
import uuid

from openstack.tests.unit import base


class TestUsage(base.TestCase):
    def test_get_usage(self):
        project = self.mock_for_keystone_projects(
            project_count=1, id_get=True
        )[0]
        start = end = datetime.datetime.now()

        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-simple-tenant-usage', project.project_id],
                        qs_elements=[
                            f'start={start.isoformat()}',
                            f'end={end.isoformat()}',
                        ],
                    ),
                    json={
                        "tenant_usage": {
                            "server_usages": [
                                {
                                    "ended_at": None,
                                    "flavor": "m1.tiny",
                                    "hours": 1.0,
                                    "instance_id": uuid.uuid4().hex,
                                    "local_gb": 1,
                                    "memory_mb": 512,
                                    "name": "instance-2",
                                    "started_at": "2012-10-08T20:10:44.541277",
                                    "state": "active",
                                    "tenant_id": "6f70656e737461636b20342065766572",  # noqa: E501
                                    "uptime": 3600,
                                    "vcpus": 1,
                                }
                            ],
                            "start": "2012-10-08T20:10:44.587336",
                            "stop": "2012-10-08T21:10:44.587336",
                            "tenant_id": "6f70656e737461636b20342065766572",
                            "total_hours": 1.0,
                            "total_local_gb_usage": 1.0,
                            "total_memory_mb_usage": 512.0,
                            "total_vcpus_usage": 1.0,
                        }
                    },
                ),
            ]
        )

        self.cloud.get_compute_usage(project.project_id, start, end)

        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from typing import Any, cast

import requests

from keystoneauth1 import adapter

from openstack.common import metadata
from openstack import format
from openstack import resource
from openstack import utils


class Volume(resource.Resource, metadata.MetadataMixin):
    resource_key = "volume"
    resources_key = "volumes"
    base_path = "/volumes"

    _query_mapping = resource.QueryParameters(
        'name', 'status', 'project_id', all_projects='all_tenants'
    )

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_commit = True
    allow_list = True

    # Properties
    #: TODO(briancurtin): This is currently undocumented in the API.
    attachments = resource.Body("attachments")
    #: The availability zone.
    availability_zone = resource.Body("availability_zone")
    #: ID of the consistency group.
    consistency_group_id = resource.Body("consistencygroup_id")
    #: The timestamp of this volume creation.
    created_at = resource.Body("created_at")
    #: The volume description.
    description = resource.Body("description")
    #: Extended replication status on this volume.
    extended_replication_status = resource.Body(
        "os-volume-replication:extended_status"
    )
    #: The volume's current back-end.
    host = resource.Body("os-vol-host-attr:host")
    #: The ID of the image from which you want to create the volume.
    #: Required to create a bootable volume.
    image_id = resource.Body("imageRef")
    #: Enables or disables the bootable attribute. You can boot an
    #: instance from a bootable volume. *Type: bool*
    is_bootable = resource.Body("bootable", type=format.BoolStr)
    #: ``True`` if this volume is encrypted, ``False`` if not.
    #: *Type: bool*
    is_encrypted = resource.Body("encrypted", type=format.BoolStr)
    #: Whether volume will be sharable or not.
    is_multiattach = resource.Body("multiattach", type=format.BoolStr)
    #: The volume ID that this volume's name on the back-end is based on.
    migration_id = resource.Body("os-vol-mig-status-attr:name_id")
    #: The status of this volume's migration (None means that a migration
    #: is not currently in progress).
    migration_status = resource.Body("os-vol-mig-status-attr:migstat")
    #: The project ID associated with current back-end.
    project_id = resource.Body("os-vol-tenant-attr:tenant_id")
    #: Data set by the replication driver
    replication_driver_data = resource.Body(
        "os-volume-replication:driver_data"
    )
    #: Status of replication on this volume.
    replication_status = resource.Body("replication_status")
    #: Scheduler hints for the volume
    scheduler_hints = resource.Body('OS-SCH-HNT:scheduler_hints', type=dict)
    #: The size of the volume, in gibibytes (GiB). *Type: int*
    size = resource.Body("size", type=int)
    #: To create a volume from an existing snapshot, specify the ID of
    #: the existing volume snapshot. If specified, the volume is created
    #: in same availability zone and with same size of the snapshot.
    snapshot_id = resource.Body("snapshot_id")
    #: To create a volume from an existing volume, specify the ID of
    #: the existing volume. If specified, the volume is created with
    #: same size of the source volume.
    source_volume_id = resource.Body("source_volid")
    #: One of the following values: creating, available, attaching, in-use
    #: deleting, error, error_deleting, backing-up, restoring-backup,
    #: error_restoring. For details on these statuses, see the
    #: Block Storage API documentation.
    status = resource.Body("status")
    #: The date and time when the resource was updated.
    updated_at = resource.Body("updated_at")
    #: The user ID associated with the volume
    user_id = resource.Body("user_id")
    #: One or more metadata key and value pairs about image
    volume_image_metadata = resource.Body("volume_image_metadata")
    #: The name of the associated volume type.
    volume_type = resource.Body("volume_type")

    def _action(
        self, session: adapter.Adapter, body: dict[str, Any]
    ) -> requests.Response:
        """Preform volume actions given the message body."""
        # NOTE: This is using Volume.base_path instead of self.base_path
        # as both Volume and VolumeDetail instances can be acted on, but
        # the URL used is sans any additional /detail/ part.
        url = utils.urljoin(Volume.base_path, self.id, 'action')
        return session.post(url, json=body)

    def extend(self, session: adapter.Adapter, size: int) -> None:
        """Extend a volume size."""
        body = {'os-extend': {'new_size': size}}
        self._action(session, body)

    def set_bootable_status(
        self, session: adapter.Adapter, bootable: bool = True
    ) -> None:
        """Set volume bootable status flag"""
        body = {'os-set_bootable': {'bootable': bootable}}
        self._action(session, body)

    def set_readonly(self, session: adapter.Adapter, readonly: bool) -> None:
        """Set volume readonly flag"""
        body = {'os-update_readonly_flag': {'readonly': readonly}}
        self._action(session, body)

    def set_image_metadata(
        self, session: adapter.Adapter, metadata: dict[str, str]
    ) -> None:
        """Sets image metadata key-value pairs on the volume"""
        body = {'os-set_image_metadata': {'metadata': metadata}}
        self._action(session, body)

    def delete_image_metadata(self, session: adapter.Adapter) -> None:
        """Remove all image metadata from the volume"""
        for key in self.metadata:
            body = {'os-unset_image_metadata': {'key': key}}
            self._action(session, body)

    def delete_image_metadata_item(
        self, session: adapter.Adapter, key: str
    ) -> None:
        """Remove a single image metadata from the volume"""
        body = {'os-unset_image_metadata': {'key': key}}
        self._action(session, body)

    def reset_status(
        self,
        session: adapter.Adapter,
        status: str | None = None,
        attach_status: str | None = None,
        migration_status: str | None = None,
    ) -> None:
        """Reset volume statuses (admin operation)"""
        body: dict[str, dict[str, str]] = {'os-reset_status': {}}
        if status:
            body['os-reset_status']['status'] = status
        if attach_status:
            body['os-reset_status']['attach_status'] = attach_status
        if migration_status:
            body['os-reset_status']['migration_status'] = migration_status
        self._action(session, body)

    def attach(
        self,
        session: adapter.Adapter,
        mountpoint: str,
        instance: str | None = None,
        host_name: str | None = None,
    ) -> None:
        """Attach volume to server"""
        body = {
            'os-attach': {'mountpoint': mountpoint},
        }
        if instance is not None:
            body['os-attach']['instance_uuid'] = instance
        elif host_name is not None:
            body['os-attach']['host_name'] = host_name
        else:
            raise ValueError(
                'Either instance_uuid or host_name must be specified'
            )

        self._action(session, body)

    def detach(
        self,
        session: adapter.Adapter,
        attachment: str,
        force: bool = False,
        connector: dict[str, Any] | None = None,
    ) -> None:
        """Detach volume from server"""
        if not force:
            body = {'os-detach': {'attachment_id': attachment}}
        if force:
            inner: dict[str, Any] = {'attachment_id': attachment}
            if connector:
                inner['connector'] = connector
            body = {'os-force_detach': inner}

        self._action(session, body)

    def unmanage(self, session: adapter.Adapter) -> None:
        """Unmanage volume"""
        body = {'os-unmanage': None}

        self._action(session, body)

    def retype(
        self,
        session: adapter.Adapter,
        new_type: str,
        migration_policy: str | None = None,
    ) -> None:
        """Change volume type"""
        body = {'os-retype': {'new_type': new_type}}
        if migration_policy:
            body['os-retype']['migration_policy'] = migration_policy

        self._action(session, body)

    def migrate(
        self,
        session: adapter.Adapter,
        host: str | None = None,
        force_host_copy: bool = False,
        lock_volume: bool = False,
    ) -> None:
        """Migrate volume"""
        req: dict[str, Any] = {}
        if host is not None:
            req['host'] = host
        if force_host_copy:
            req['force_host_copy'] = force_host_copy
        if lock_volume:
            req['lock_volume'] = lock_volume
        body = {'os-migrate_volume': req}

        self._action(session, body)

    def complete_migration(
        self,
        session: adapter.Adapter,
        new_volume_id: str,
        error: bool = False,
    ) -> None:
        """Complete volume migration"""
        body = {
            'os-migrate_volume_completion': {
                'new_volume': new_volume_id,
                'error': error,
            }
        }

        self._action(session, body)

    def force_delete(self, session: adapter.Adapter) -> None:
        """Force volume deletion"""
        body = {'os-force_delete': None}

        self._action(session, body)

    def upload_to_image(
        self,
        session: adapter.Adapter,
        image_name: str,
        *,
        force: bool = False,
        disk_format: str | None = None,
        container_format: str | None = None,
    ) -> dict[str, Any]:
        """Upload the volume to image service"""
        body = {
            'os-volume_upload_image': {
                'image_name': image_name,
                'force': force,
                'disk_format': disk_format,
                'container_format': container_format,
            },
        }

        resp = cast(dict[str, Any], self._action(session, body).json())
        return cast(dict[str, Any], resp['os-volume_upload_image'])


VolumeDetail = Volume



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.network.v2 import network as _network
from openstack.network.v2 import port as _port
from openstack.network.v2 import tap_flow as _tap_flow
from openstack.network.v2 import tap_service as _tap_service
from openstack.tests.functional import base


class TestTapService(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        if not self.user_cloud.network.find_extension("taas"):
            self.skipTest("Neutron Tap-as-a-service Extension disabled")

        self.TAP_S_NAME = 'my_service' + self.getUniqueString()
        self.TAP_F_NAME = 'my_flow' + self.getUniqueString()
        net = self.user_cloud.network.create_network()
        assert isinstance(net, _network.Network)
        self.SERVICE_NET_ID = net.id

        net = self.user_cloud.network.create_network()
        assert isinstance(net, _network.Network)
        self.FLOW_NET_ID = net.id

        port = self.user_cloud.network.create_port(
            network_id=self.SERVICE_NET_ID
        )
        assert isinstance(port, _port.Port)
        self.SERVICE_PORT_ID = port.id

        port = self.user_cloud.network.create_port(network_id=self.FLOW_NET_ID)
        assert isinstance(port, _port.Port)
        self.FLOW_PORT_ID = port.id

        tap_service = self.user_cloud.network.create_tap_service(
            name=self.TAP_S_NAME, port_id=self.SERVICE_PORT_ID
        )
        assert isinstance(tap_service, _tap_service.TapService)
        self.TAP_SERVICE = tap_service

        tap_flow = self.user_cloud.network.create_tap_flow(
            name=self.TAP_F_NAME,
            tap_service_id=self.TAP_SERVICE.id,
            source_port=self.FLOW_PORT_ID,
            direction='BOTH',
        )
        assert isinstance(tap_flow, _tap_flow.TapFlow)
        self.TAP_FLOW = tap_flow

    def tearDown(self):
        sot = self.user_cloud.network.delete_tap_flow(
            self.TAP_FLOW.id, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_tap_service(
            self.TAP_SERVICE.id, ignore_missing=False
        )
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_port(self.SERVICE_PORT_ID)
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_port(self.FLOW_PORT_ID)
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_network(self.SERVICE_NET_ID)
        self.assertIsNone(sot)
        sot = self.user_cloud.network.delete_network(self.FLOW_NET_ID)
        self.assertIsNone(sot)
        super().tearDown()

    def test_find_tap_service(self):
        sot = self.user_cloud.network.find_tap_service(
            self.TAP_SERVICE.name, ignore_missing=False
        )
        self.assertEqual(self.SERVICE_PORT_ID, sot.port_id)
        self.assertEqual(self.TAP_S_NAME, sot.name)

    def test_get_tap_service(self):
        sot = self.user_cloud.network.get_tap_service(self.TAP_SERVICE.id)
        self.assertEqual(self.SERVICE_PORT_ID, sot.port_id)
        self.assertEqual(self.TAP_S_NAME, sot.name)

    def test_list_tap_services(self):
        tap_service_ids = [
            ts.id for ts in self.user_cloud.network.tap_services()
        ]
        self.assertIn(self.TAP_SERVICE.id, tap_service_ids)

    def test_update_tap_service(self):
        description = 'My tap service'
        sot = self.user_cloud.network.update_tap_service(
            self.TAP_SERVICE.id, description=description
        )
        self.assertEqual(description, sot.description)

    def test_find_tap_flow(self):
        sot = self.user_cloud.network.find_tap_flow(
            self.TAP_FLOW.name, ignore_missing=False
        )
        self.assertEqual(self.FLOW_PORT_ID, sot.source_port)
        self.assertEqual(self.TAP_SERVICE.id, sot.tap_service_id)
        self.assertEqual('BOTH', sot.direction)
        self.assertEqual(self.TAP_F_NAME, sot.name)

    def test_get_tap_flow(self):
        sot = self.user_cloud.network.get_tap_flow(self.TAP_FLOW.id)
        self.assertEqual(self.FLOW_PORT_ID, sot.source_port)
        self.assertEqual(self.TAP_F_NAME, sot.name)
        self.assertEqual(self.TAP_SERVICE.id, sot.tap_service_id)
        self.assertEqual('BOTH', sot.direction)

    def test_list_tap_flows(self):
        tap_flow_ids = [tf.id for tf in self.user_cloud.network.tap_flows()]
        self.assertIn(self.TAP_FLOW.id, tap_flow_ids)

    def test_update_tap_flow(self):
        description = 'My tap flow'
        sot = self.user_cloud.network.update_tap_flow(
            self.TAP_FLOW.id, description=description
        )
        self.assertEqual(description, sot.description)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings
from typing import Any, Self

import requests

from keystoneauth1 import adapter

from openstack.common import metadata
from openstack import exceptions
from openstack import format
from openstack import resource
from openstack import utils
from openstack import warnings as os_warnings


class Snapshot(resource.Resource, metadata.MetadataMixin):
    resource_key = "snapshot"
    resources_key = "snapshots"
    base_path = "/snapshots"

    _query_mapping = resource.QueryParameters(
        'name', 'status', 'volume_id', all_projects='all_tenants'
    )

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_commit = True
    allow_list = True

    # Properties
    #: The timestamp of this snapshot creation.
    created_at = resource.Body("created_at")
    #: Description of snapshot. Default is None.
    description = resource.Body("description")
    #: Indicate whether to create snapshot, even if the volume is attached.
    #: Default is ``False``. *Type: bool*
    is_forced = resource.Body("force", type=format.BoolStr)
    #: The size of the volume, in gibibytes (GiB).
    size = resource.Body("size", type=int)
    #: The current status of this snapshot. Potential values are creating,
    #: available, deleting, error, and error_deleting.
    status = resource.Body("status")
    #: The date and time when the resource was updated.
    updated_at = resource.Body("updated_at")
    #: The ID of the volume this snapshot was taken of.
    volume_id = resource.Body("volume_id")

    def _action(
        self, session: adapter.Adapter, body: dict[str, Any]
    ) -> requests.Response:
        """Preform backup actions given the message body."""
        url = utils.urljoin(self.base_path, self.id, 'action')
        resp = session.post(url, json=body)
        exceptions.raise_from_response(resp)
        return resp

    def force_delete(self, session: adapter.Adapter) -> None:
        """Force snapshot deletion."""
        body = {'os-force_delete': None}
        self._action(session, body)

    def reset_status(self, session: adapter.Adapter, status: str) -> None:
        """Reset the status of the snapshot."""
        body = {'os-reset_status': {'status': status}}
        self._action(session, body)

    def reset(self, session: adapter.Adapter, status: str) -> None:
        warnings.warn(
            "reset is a deprecated alias for reset_status and will be "
            "removed in a future release.",
            os_warnings.RemovedInSDK60Warning,
        )
        self.reset_status(session, status)

    @classmethod
    def manage(
        cls,
        session: adapter.Adapter,
        volume_id: str,
        ref: dict[str, Any],
        name: str | None = None,
        description: str | None = None,
        metadata: dict[str, str] | None = None,
    ) -> Self:
        """Manage a snapshot under block storage provisioning."""
        url = '/os-snapshot-manage'
        body = {
            'snapshot': {
                'volume_id': volume_id,
                'ref': ref,
                'name': name,
                'description': description,
                'metadata': metadata,
            }
        }
        resp = session.post(url, json=body)
        exceptions.raise_from_response(resp)
        snapshot = cls()
        snapshot._translate_response(resp)
        return snapshot

    def unmanage(self, session: adapter.Adapter) -> None:
        """Unmanage a snapshot from block storage provisioning."""
        body = {'os-unmanage': None}
        self._action(session, body)


SnapshotDetail = Snapshot



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# Copyright 2014 OpenStack Foundation
# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools
import os
import pprint
import select
import socket
import threading
import time

import fixtures
from keystoneauth1 import exceptions
import prometheus_client
from requests import exceptions as rexceptions
import testtools.content

from openstack.tests.unit import base


class StatsdFixture(fixtures.Fixture):
    def _setUp(self):
        self.running = True
        self.thread = threading.Thread(target=self.run)
        self.thread.daemon = True
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(('', 0))
        self.port = self.sock.getsockname()[1]
        self.wake_read, self.wake_write = os.pipe()
        self.stats = []
        self.thread.start()
        self.addCleanup(self._cleanup)

    def run(self):
        while self.running:
            poll = select.poll()
            poll.register(self.sock, select.POLLIN)
            poll.register(self.wake_read, select.POLLIN)
            ret = poll.poll()
            for fd, event in ret:
                if fd == self.sock.fileno():
                    data = self.sock.recvfrom(1024)
                    if not data:
                        return
                    self.stats.append(data[0])
                if fd == self.wake_read:
                    return

    def _cleanup(self):
        self.running = False
        os.write(self.wake_write, b'1\n')
        self.thread.join()


class TestStats(base.TestCase):
    def setUp(self):
        self.statsd = StatsdFixture()
        self.useFixture(self.statsd)
        # note, use 127.0.0.1 rather than localhost to avoid getting ipv6
        # see: https://github.com/jsocol/pystatsd/issues/61
        self.useFixture(
            fixtures.EnvironmentVariable('STATSD_HOST', '127.0.0.1')
        )
        self.useFixture(
            fixtures.EnvironmentVariable('STATSD_PORT', str(self.statsd.port))
        )

        self.add_info_on_exception('statsd_content', self.statsd.stats)
        # Set up the above things before the super setup so that we have the
        # environment variables set when the Connection is created.
        super().setUp()

        self._registry = prometheus_client.CollectorRegistry()
        self.cloud.config._collector_registry = self._registry
        self.addOnException(self._add_prometheus_samples)

    def _add_prometheus_samples(self, exc_info):
        samples = []
        for metric in self._registry.collect():
            for s in metric.samples:
                samples.append(s)
        self.addDetail(
            'prometheus_samples',
            testtools.content.text_content(pprint.pformat(samples)),
        )

    def assert_reported_stat(self, key, value=None, kind=None):
        """Check statsd output

        Check statsd return values.  A ``value`` should specify a
        ``kind``, however a ``kind`` may be specified without a
        ``value`` for a generic match.  Leave both empty to just check
        for key presence.

        :arg str key: The statsd key
        :arg str value: The expected value of the metric ``key``
        :arg str kind: The expected type of the metric ``key``  For example

          - ``c`` counter
          - ``g`` gauge
          - ``ms`` timing
          - ``s`` set

        Note that for ``ms`` type, you are expressing a maximum value,
        not an exact value. This is to avoid flakey tests.
        """

        self.assertIsNotNone(self.statsd)

        if value:
            self.assertNotEqual(kind, None)

        start = time.time()
        while time.time() < (start + 1):
            # Note our fake statsd just queues up results in a queue.
            # We just keep going through them until we find one that
            # matches, or fail out.  If statsd pipelines are used,
            # large single packets are sent with stats separated by
            # newlines; thus we first flatten the stats out into
            # single entries.
            stats = itertools.chain.from_iterable(
                [s.decode('utf-8').split('\n') for s in self.statsd.stats]
            )
            for stat in stats:
                k, v = stat.split(':')
                if key == k:
                    if kind is None:
                        # key with no qualifiers is found
                        return True

                    s_value, s_kind = v.split('|')

                    # if no kind match, look for other keys
                    if kind != s_kind:
                        continue

                    if value:
                        # special-case value|ms because statsd can turn
                        # timing results into float of indeterminate
                        # length, hence foiling string matching.
                        if kind == 'ms':
                            if float(value) >= float(s_value):
                                return True
                        elif value == s_value:
                            return True
                        # otherwise keep looking for other matches
                        continue

                    # this key matches
                    return True
            time.sleep(0.1)

        raise Exception(f"Key {key} not found in reported stats")

    def assert_prometheus_stat(self, name, value, labels=None):
        sample_value = self._registry.get_sample_value(name, labels)
        self.assertEqual(sample_value, value)

    def test_list_projects(self):
        mock_uri = self.get_mock_url(
            service_type='identity', resource='projects', base_url_append='v3'
        )

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=mock_uri,
                    status_code=200,
                    json={'projects': []},
                )
            ]
        )

        self.cloud.list_projects()
        self.assert_calls()

        self.assert_reported_stat(
            'openstack.api.identity.GET.projects.200', value='1', kind='c'
        )
        self.assert_prometheus_stat(
            'openstack_http_requests_total',
            1,
            dict(
                service_type='identity',
                endpoint=mock_uri,
                method='GET',
                status_code='200',
            ),
        )

    def test_projects(self):
        mock_uri = self.get_mock_url(
            service_type='identity', resource='projects', base_url_append='v3'
        )

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=mock_uri,
                    status_code=200,
                    json={'projects': []},
                )
            ]
        )

        list(self.cloud.identity.projects())
        self.assert_calls()

        self.assert_reported_stat(
            'openstack.api.identity.GET.projects.200', value='1', kind='c'
        )
        self.assert_prometheus_stat(
            'openstack_http_requests_total',
            1,
            dict(
                service_type='identity',
                endpoint=mock_uri,
                method='GET',
                status_code='200',
            ),
        )

    def test_servers(self):
        mock_uri = 'https://compute.example.com/v2.1/servers/detail'

        self.register_uris(
            [
                self.get_nova_discovery_mock_dict(),
                dict(
                    method='GET',
                    uri=mock_uri,
                    status_code=200,
                    json={'servers': []},
                ),
            ]
        )

        list(self.cloud.compute.servers())
        self.assert_calls()

        self.assert_reported_stat(
            'openstack.api.compute.GET.servers_detail.200', value='1', kind='c'
        )
        self.assert_reported_stat(
            'openstack.api.compute.GET.servers_detail.200',
            value='5',
            kind='ms',
        )
        self.assert_prometheus_stat(
            'openstack_http_requests_total',
            1,
            dict(
                service_type='compute',
                endpoint=mock_uri,
                method='GET',
                status_code='200',
            ),
        )

    def test_servers_no_detail(self):
        mock_uri = 'https://compute.example.com/v2.1/servers'

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=mock_uri,
                    status_code=200,
                    json={'servers': []},
                )
            ]
        )

        self.cloud.compute.get('/servers')
        self.assert_calls()

        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.200', value='1', kind='c'
        )
        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.200', value='5', kind='ms'
        )
        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.attempted', value='1', kind='c'
        )
        self.assert_prometheus_stat(
            'openstack_http_requests_total',
            1,
            dict(
                service_type='compute',
                endpoint=mock_uri,
                method='GET',
                status_code='200',
            ),
        )

    def test_servers_error(self):
        mock_uri = 'https://compute.example.com/v2.1/servers'

        self.register_uris(
            [dict(method='GET', uri=mock_uri, status_code=500, json={})]
        )

        self.cloud.compute.get('/servers')
        self.assert_calls()

        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.500', value='1', kind='c'
        )
        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.500', value='5', kind='ms'
        )
        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.attempted', value='1', kind='c'
        )
        self.assert_prometheus_stat(
            'openstack_http_requests_total',
            1,
            dict(
                service_type='compute',
                endpoint=mock_uri,
                method='GET',
                status_code='500',
            ),
        )

    def test_timeout(self):
        mock_uri = 'https://compute.example.com/v2.1/servers'

        self.register_uris(
            [dict(method='GET', uri=mock_uri, exc=rexceptions.ConnectTimeout)]
        )

        try:
            self.cloud.compute.get('/servers')
        except exceptions.ConnectTimeout:
            pass

        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.failed', value='1', kind='c'
        )
        self.assert_reported_stat(
            'openstack.api.compute.GET.servers.attempted', value='1', kind='c'
        )


class TestNoStats(base.TestCase):
    def setUp(self):
        super().setUp()
        self.statsd = StatsdFixture()
        self.useFixture(self.statsd)

    def test_no_stats(self):
        mock_uri = self.get_mock_url(
            service_type='identity', resource='projects', base_url_append='v3'
        )

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=mock_uri,
                    status_code=200,
                    json={'projects': []},
                )
            ]
        )

        self.cloud.identity._statsd_client = None
        list(self.cloud.identity.projects())
        self.assert_calls()
        self.assertEqual([], self.statsd.stats)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.key_manager.v1 import quota
from openstack.tests.unit import base


EXAMPLE = {
    'secrets': 10,
    'orders': 20,
    'containers': 10,
    'consumers': -1,
    'cas': 5,
}


class TestQuota(base.TestCase):
    def test_basic(self):
        sot = quota.Quota()
        self.assertEqual('quotas', sot.resource_key)
        self.assertEqual('quotas', sot.resources_key)
        self.assertEqual('/quotas', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertFalse(sot.allow_commit)
        self.assertFalse(sot.allow_delete)
        self.assertFalse(sot.allow_list)

    def test_make_it(self):
        sot = quota.Quota(**EXAMPLE)
        self.assertEqual(EXAMPLE['secrets'], sot.secrets)
        self.assertEqual(EXAMPLE['orders'], sot.orders)
        self.assertEqual(EXAMPLE['containers'], sot.containers)
        self.assertEqual(EXAMPLE['consumers'], sot.consumers)
        self.assertEqual(EXAMPLE['cas'], sot.cas)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack.tests.functional.identity.v3 import base


class TestAccessRule(base.BaseIdentityTest):
    def setUp(self):
        super().setUp()

        if not self.operator_cloud:
            self.skipTest("Operator cloud is required for this test")

        user_id = self.operator_cloud.current_user_id
        self.assertIsNotNone(user_id)
        assert user_id is not None
        self.user_id = user_id

    def _create_application_credential_with_access_rule(self):
        """create application credential with access_rule."""

        assert self.user_id is not None
        app_cred = self.admin_identity_client.create_application_credential(
            user=self.user_id,
            name='app_cred',
            access_rules=[
                {
                    "path": "/v2.0/metrics",
                    "service": "monitoring",
                    "method": "GET",
                }
            ],
        )
        self.addCleanup(
            self.admin_identity_client.delete_application_credential,
            self.user_id,
            app_cred['id'],
        )
        return app_cred

    def test_get_access_rule(self):
        app_cred = self._create_application_credential_with_access_rule()
        access_rule_id = app_cred['access_rules'][0]['id']
        access_rule = self.admin_identity_client.get_access_rule(
            user=self.user_id, access_rule=access_rule_id
        )
        self.assertEqual(access_rule['id'], access_rule_id)
        self.assertEqual(access_rule['user_id'], self.user_id)

    def test_list_access_rules(self):
        app_cred = self._create_application_credential_with_access_rule()
        access_rule_id = app_cred['access_rules'][0]['id']
        access_rules = self.admin_identity_client.access_rules(
            user=self.user_id
        )
        self.assertEqual(1, len(list(access_rules)))
        for access_rule in access_rules:
            self.assertEqual(app_cred['user_id'], self.user_id)
            self.assertEqual(access_rule_id, access_rule['id'])

    def test_delete_access_rule(self):
        app_cred = self._create_application_credential_with_access_rule()
        access_rule_id = app_cred['access_rules'][0]['id']

        # This is expected to raise an exception since access_rule is still
        # in use for app_cred.
        self.assertRaises(
            exceptions.HttpException,
            self.admin_identity_client.delete_access_rule,
            user=self.user_id,
            access_rule=access_rule_id,
        )

        # delete application credential first to delete access rule
        self.admin_identity_client.delete_application_credential(
            user=self.user_id, application_credential=app_cred['id']
        )
        # delete orphaned access rules
        self.admin_identity_client.delete_access_rule(
            user=self.user_id, access_rule=access_rule_id
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v2 import snapshot as _snapshot
from openstack.block_storage.v2 import volume as _volume
from openstack.tests.functional.block_storage.v2 import base
from openstack import utils as sdk_utils


class TestSnapshot(base.BaseBlockStorageTest):
    def setUp(self):
        super().setUp()

        self.SNAPSHOT_NAME = self.getUniqueString()
        self.VOLUME_NAME = self.getUniqueString()

        volume = self.user_cloud.block_storage.create_volume(
            name=self.VOLUME_NAME, size=1
        )
        self.user_cloud.block_storage.wait_for_status(
            volume,
            status='available',
            failures=['error'],
            interval=2,
            wait=self._wait_for_timeout,
        )
        assert isinstance(volume, _volume.Volume)
        self.assertEqual(self.VOLUME_NAME, volume.name)
        self.VOLUME_ID = volume.id
        snapshot = self.user_cloud.block_storage.create_snapshot(
            name=self.SNAPSHOT_NAME, volume_id=self.VOLUME_ID
        )
        self.user_cloud.block_storage.wait_for_status(
            snapshot,
            status='available',
            failures=['error'],
            interval=2,
            wait=self._wait_for_timeout,
        )
        assert isinstance(snapshot, _snapshot.Snapshot)
        self.assertEqual(self.SNAPSHOT_NAME, snapshot.name)
        self.SNAPSHOT_ID = snapshot.id

    def tearDown(self):
        block_storage_client = sdk_utils.ensure_service_version(
            self.user_cloud.block_storage, '2'
        )
        snapshot = block_storage_client.get_snapshot(self.SNAPSHOT_ID)
        sot = block_storage_client.delete_snapshot(
            snapshot, ignore_missing=False
        )
        block_storage_client.wait_for_delete(
            snapshot, interval=2, wait=self._wait_for_timeout
        )
        self.assertIsNone(sot)
        sot = block_storage_client.delete_volume(
            self.VOLUME_ID, ignore_missing=False
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_get(self):
        sot = self.user_cloud.block_storage.get_snapshot(self.SNAPSHOT_ID)
        self.assertEqual(self.SNAPSHOT_NAME, sot.name)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import identity_provider
from openstack.tests.unit import base


IDENTIFIER = 'IDENTIFIER'
EXAMPLE = {
    'id': IDENTIFIER,
    'domain_id': 'example_domain',
    'description': 'An example description',
    'is_enabled': True,
    'remote_ids': ['https://auth.example.com/auth/realms/ExampleRealm'],
    'authorization_ttl': 7,
}


class TestIdentityProvider(base.TestCase):
    def test_basic(self):
        sot = identity_provider.IdentityProvider()
        self.assertEqual('identity_provider', sot.resource_key)
        self.assertEqual('identity_providers', sot.resources_key)
        self.assertEqual('/OS-FEDERATION/identity_providers', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)
        self.assertEqual('PUT', sot.create_opts.method)

        self.assertDictEqual(
            {
                'id': 'id',
                'limit': 'limit',
                'marker': 'marker',
                'is_enabled': 'enabled',
            },
            sot._query_mapping._mapping,
        )

    def test_make_it(self):
        sot = identity_provider.IdentityProvider(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['id'], sot.name)
        self.assertEqual(EXAMPLE['domain_id'], sot.domain_id)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['is_enabled'], sot.is_enabled)
        self.assertEqual(EXAMPLE['remote_ids'], sot.remote_ids)
        self.assertEqual(EXAMPLE['authorization_ttl'], sot.authorization_ttl)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Literal, Self, overload

from keystoneauth1 import adapter

from openstack.common import tag
from openstack import exceptions
from openstack import resource
from openstack import utils


class Stack(resource.Resource):
    name_attribute = 'stack_name'
    resource_key = 'stack'
    resources_key = 'stacks'
    base_path = '/stacks'

    # capabilities
    allow_create = True
    allow_list = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True

    create_opts = resource.CreateOpts(request_key=None)

    _query_mapping = resource.QueryParameters(
        'action',
        'name',
        'status',
        'project_id',
        'owner_id',
        'username',
        project_id='tenant_id',
        **tag.TagMixin._tag_query_parameters,
    )

    # Properties
    #: A list of resource objects that will be added if a stack update
    #  is performed.
    added = resource.Body('added')
    #: Placeholder for AWS compatible template listing capabilities
    #: required by the stack.
    capabilities = resource.Body('capabilities')
    #: Timestamp of the stack creation.
    created_at = resource.Body('creation_time')
    #: A text description of the stack.
    description = resource.Body('description')
    #: A list of resource objects that will be deleted if a stack
    #: update is performed.
    deleted = resource.Body('deleted', type=list)
    #: Timestamp of the stack deletion.
    deleted_at = resource.Body('deletion_time')
    #: A JSON environment for the stack.
    environment = resource.Body('environment')
    #: An ordered list of names for environment files found in the files dict.
    environment_files = resource.Body('environment_files', type=list)
    #: Additional files referenced in the template or the environment
    files = resource.Body('files', type=dict)
    #: Name of the container in swift that has child
    #: templates and environment files.
    files_container = resource.Body('files_container')
    #: Whether the stack will support a rollback operation on stack
    #: create/update failures. *Type: bool*
    is_rollback_disabled = resource.Body('disable_rollback', type=bool)
    #: A list of dictionaries containing links relevant to the stack.
    links = resource.Body('links')
    #: Name of the stack.
    name = resource.Body('stack_name')
    stack_name = resource.URI('stack_name')
    #: Placeholder for future extensions where stack related events
    #: can be published.
    notification_topics = resource.Body('notification_topics')
    #: A list containing output keys and values from the stack, if any.
    outputs = resource.Body('outputs')
    #: The ID of the owner stack if any.
    owner_id = resource.Body('stack_owner')
    #: A dictionary containing the parameter names and values for the stack.
    parameters = resource.Body('parameters', type=dict)
    #: The ID of the parent stack if any
    parent_id = resource.Body('parent')
    #: A list of resource objects that will be replaced if a stack update
    #: is performed.
    replaced = resource.Body('replaced')
    #: A string representation of the stack status, e.g. ``CREATE_COMPLETE``.
    status = resource.Body('stack_status')
    #: A text explaining how the stack transits to its current status.
    status_reason = resource.Body('stack_status_reason')
    #: A list of strings used as tags on the stack
    tags = resource.Body('tags', type=list, default=[])
    #: A dict containing the template use for stack creation.
    template = resource.Body('template', type=dict)
    #: Stack template description text. Currently contains the same text
    #: as that of the ``description`` property.
    template_description = resource.Body('template_description')
    #: A string containing the URL where a stack template can be found.
    template_url = resource.Body('template_url')
    #: Stack operation timeout in minutes.
    timeout_mins = resource.Body('timeout_mins')
    #: A list of resource objects that will remain unchanged if a stack
    #: update is performed.
    unchanged = resource.Body('unchanged')
    #: A list of resource objects that will have their properties updated
    #: in place if a stack update is performed.
    updated = resource.Body('updated')
    #: Timestamp of last update on the stack.
    updated_at = resource.Body('updated_time')
    #: The ID of the user project created for this stack.
    user_project_id = resource.Body('stack_user_project_id')

    def commit(
        self,
        session: adapter.Adapter,
        prepend_key: bool = True,
        has_body: bool = True,
        retry_on_conflict: bool | None = None,
        base_path: str | None = None,
        *,
        microversion: str | None = None,
        preview: bool = False,
        **kwargs: Any,
    ) -> Self:
        # This overrides the default behavior of resource update because
        # we need to use other endpoint for update preview.
        self._body._dirty.discard("id")
        base_path = None
        if self.name and self.id:
            base_path = f'/stacks/{self.name}/{self.id}'
        elif self.name or self.id:
            # We have only one of name/id. Do not try to build a stacks/NAME/ID
            # path
            base_path = f'/stacks/{self.name or self.id}'
        request = self._prepare_request(
            prepend_key=False, requires_id=False, base_path=base_path
        )

        microversion = self._get_microversion(session)

        request_url = request.url
        if preview:
            request_url = utils.urljoin(request_url, 'preview')

        response = session.put(
            request_url,
            json=request.body,
            headers=request.headers,
            microversion=microversion,
        )

        self.microversion = microversion
        self._translate_response(response, has_body=True)
        return self

    def _action(self, session: adapter.Adapter, body: dict[str, Any]) -> Any:
        """Perform stack actions"""
        url = utils.urljoin(self.base_path, self._get_id(self), 'actions')
        resp = session.post(url, json=body, microversion=self.microversion)
        exceptions.raise_from_response(resp)
        return resp

    def check(self, session: adapter.Adapter) -> Any:
        return self._action(session, {'check': ''})

    def abandon(self, session: adapter.Adapter) -> dict[str, Any]:
        url = utils.urljoin(
            self.base_path, self.name, self._get_id(self), 'abandon'
        )
        resp = session.delete(url)
        return resp.json()  # type: ignore[no-any-return]

    def export(self, session: adapter.Adapter) -> dict[str, Any]:
        """Export a stack data

        :param session: The session to use for making this request.
        :return: A dictionary containing the stack data.
        """
        url = utils.urljoin(
            self.base_path, self.name, self._get_id(self), 'export'
        )
        resp = session.get(url)
        exceptions.raise_from_response(resp)
        return resp.json()  # type: ignore[no-any-return]

    def suspend(self, session: adapter.Adapter) -> None:
        """Suspend a stack

        :param session: The session to use for making this request
        :returns: None
        """
        body = {"suspend": None}
        self._action(session, body)

    def resume(self, session: adapter.Adapter) -> None:
        """Resume a stack

        :param session: The session to use for making this request
        :returns: None
        """
        body = {"resume": None}
        self._action(session, body)

    def fetch(
        self,
        session: adapter.Adapter,
        requires_id: bool = True,
        base_path: str | None = None,
        error_message: str | None = None,
        skip_cache: bool = False,
        *,
        resolve_outputs: bool = True,
        **params: Any,
    ) -> Self:
        if not self.allow_fetch:
            raise exceptions.MethodNotSupported(self, "fetch")

        request = self._prepare_request(
            requires_id=requires_id, base_path=base_path
        )
        # session = self._get_session(session)
        microversion = self._get_microversion(session)

        # NOTE(gtema): would be nice to simply use QueryParameters, however
        # Heat return 302 with parameters being set into URL and requests
        # apply parameters again, what results in them being set doubled
        if not resolve_outputs:
            request.url = request.url + '?resolve_outputs=False'
        response = session.get(
            request.url, microversion=microversion, skip_cache=skip_cache
        )
        kwargs: dict[str, Any] = {}
        if error_message:
            kwargs['error_message'] = error_message

        self.microversion = microversion
        self._translate_response(response, **kwargs)

        if self and self.status in ['DELETE_COMPLETE', 'ADOPT_COMPLETE']:
            raise exceptions.NotFoundException(f"No stack found for {self.id}")
        return self

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: Literal[False],
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self: ...

    @overload
    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None: ...

    @classmethod
    def find(
        cls,
        session: adapter.Adapter,
        name_or_id: str,
        ignore_missing: bool = True,
        list_base_path: str | None = None,
        *,
        microversion: str | None = None,
        all_projects: bool | None = None,
        **params: Any,
    ) -> Self | None:
        """Find a resource by its name or id.

        :param session: The session to use for making this request.
        :type session: :class:`~keystoneauth1.adapter.Adapter`
        :param name_or_id: This resource's identifier, if needed by
                           the request. The default is ``None``.
        :param bool ignore_missing: When set to ``False``
                    :class:`~openstack.exceptions.NotFoundException` will be
                    raised when the resource does not exist.
                    When set to ``True``, None will be returned when
                    attempting to find a nonexistent resource.
        :param dict params: Any additional parameters to be passed into
                            underlying methods, such as to
                            :meth:`~openstack.resource.Resource.existing`
                            in order to pass on URI parameters.

        :return: The :class:`Resource` object matching the given name or id
                 or None if nothing matches.
        :raises: :class:`openstack.exceptions.DuplicateResource` if more
                 than one resource is found for this request.
        :raises: :class:`openstack.exceptions.NotFoundException` if nothing
                 is found and ignore_missing is ``False``.
        """
        session = cls._get_session(session)
        # Try to short-circuit by looking directly for a matching ID.
        try:
            match = cls.existing(
                id=name_or_id,
                connection=session._get_connection(),  # type: ignore
                **params,
            )
            return match.fetch(session, **params)
        except exceptions.NotFoundException:
            pass

        # NOTE(gtema) we do not do list, since previous call has done this
        # for us already

        if ignore_missing:
            return None

        raise exceptions.NotFoundException(
            f"No {cls.__name__} found for {name_or_id}"
        )


StackPreview = Stack
